/*
 * Copyright (c) 1994, 1997 Vectaport Inc.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation for any purpose is hereby granted without fee, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the names of the copyright holders not be used in
 * advertising or publicity pertaining to distribution of the software
 * without specific, written prior permission.  The copyright holders make
 * no representations about the suitability of this software for any purpose.
 * It is provided "as is" without express or implied warranty.
 *
 * THE COPYRIGHT HOLDERS DISCLAIM ALL WARRANTIES WITH REGARD TO THIS
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS.
 * IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY SPECIAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING
 * FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
 * NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 * 
 */

#include <Dem/dem.h>
#include <TopoFace/fgeomobjs.h>
#include <stdio.h>
#include <string.h>

DEM::DEM() {
    r = 0;
    c = 0;
    data = 0;
    _proj = NULL;
}

DEM::~DEM() {
    if (data)
	delete [] data;
}

int DEM::rows() { return r; }
int DEM::cols() { return c; }

void DEM::allocate(int rows, int cols) {
    data = new unsigned short[rows*cols];
    r = rows;
    c = cols;
    memset(data, 0, sizeof(unsigned short) * rows * cols);
}

void DEM::value(int row, int col, unsigned short value) {
    if (row >= 0 && row < r && col >= 0 && col < c)
	data[row+r*col] = value;
}

unsigned short DEM::value(int row, int col) {
    if (row >= 0 && row < r && col >= 0 && col < c)
	return data[row+r*col];
    else
	return 0;
}

// Description:
// Compute the triangle normal from three points.
inline void TriangleComputeNormal(double v1[3], double v2[3], 
                                       double v3[3], double n[3])
{
  double length, ax, ay, az, bx, by, bz;

  // order is important!!! maintain consistency with triangle vertex order 
  ax = v3[0] - v2[0]; ay = v3[1] - v2[1]; az = v3[2] - v2[2];
  bx = v1[0] - v2[0]; by = v1[1] - v2[1]; bz = v1[2] - v2[2];

  n[0] = (ay * bz - az * by);
  n[1] = (az * bx - ax * bz);
  n[2] = (ax * by - ay * bx);

  if ( (length = sqrt((n[0]*n[0] + n[1]*n[1] + n[2]*n[2]))) != 0.0 )
    {
    n[0] /= length;
    n[1] /= length;
    n[2] /= length;
    }
  else
    {
    n[0] = 1.0;
    }
}

inline double PlaneEvaluateZ(double normal[3], double origin[3], double x[3])
{
  return (( -normal[0]*(x[0]-origin[0]) - 
           normal[1]*(x[1]-origin[1]) ) / normal[2] ) + origin[2];
}


int DEM::value_by_utmpos(double ux, double uy, double& v) {
  if (_proj) {
    UV data;
    data.u = ux;
    data.v = uy;
    data = pj_inv(data, _proj);
    return value_by_geopos(data.u * RAD_TO_DEG, data.v * RAD_TO_DEG, v);
  }
  else
    return 0;
}

int DEM::value_by_geopos(double x, double y, double& v) {
  // first do a containment test in geo coords
  FPointObj pt(x, y);
  if (grunits == arcseconds) {
    FBoxObj box(_swe, _swn, _nee, _nen);
    if (!box.Contains(pt))
      return 0;
  }
  // now find the value by interpolation
  if (grunits == arcseconds) {
#if 0
    int col = int(((x - _swe) / (_see - _swe)) * cols());
    int row = int(((y - _swn) / (_nwn - _swn)) * rows());
    double p0x, p0y, p1x, p1y;
    position(row, col, p0x, p0y);
    position(row+1, col+1, p1x, p1y);
    double param_r = double(y - p0y) / double(p1y - p0y);
    double param_s = double(x - p0x) / double(p1x - p0x);
    double w0 = (1.0 - param_r) * (1.0 - param_s);
    double w1 = param_r * (1.0 - param_s);
    double w2 = (1.0 - param_r) * param_s;
    double w3 = param_r * param_s;
    v = w0 * value(row, col) + w1 * value(row, col+1) +
      w2 * value(row+1, col) + w3 * value(row+1, col+1);
#endif
    int coli = 0;
    int rowi = 0;
    double dx, dy;
    for (coli = 0; coli < cols(); coli++) {
      position(0, coli, dx, dy);
      if (dx >= x)
	break;
    }
    for (rowi = 0; rowi < rows(); rowi++) {
      position(rowi, 0, dx, dy);
      if (dy >= y)
	break;
    }
    int col = coli - 1;
    int row = rowi - 1;

    double swx, swy, nwx, nwy, nex, ney, sex, sey;
    double pswx, pswy, pnwx, pnwy, pnex, pney, psex, psey;
    position(row, col, swx, swy);
    position(row+1, col, nwx, nwy);
    position(row+1, col+1, nex, ney);
    position(row, col+1, sex, sey);
    projected_position(row, col, pswx, pswy);
    projected_position(row+1, col, pnwx, pnwy);
    projected_position(row+1, col+1, pnex, pney);
    projected_position(row, col+1, psex, psey);
    UV data;
    double px, py;
    data.u = x * DEG_TO_RAD;
    data.v = y * DEG_TO_RAD;
    data = pj_fwd(data, _proj);
    px = data.u;
    py = data.v;
    FPointObj ppt(px, py);

    float* t1x = new float[3];
    float* t1y = new float[3];
    float* t2x = new float[3];
    float* t2y = new float[3];
    t1x[0] = pnwx;
    t1y[0] = pnwy;
    t1x[1] = pnex;
    t1y[1] = pney;
    t1x[2] = pswx;
    t1x[2] = pswy;
    FFillPolygonObj tri1(t1x, t1y, 3);
    FLineObj line1_1(pnwx, pnwy, pnex, pney);
    FLineObj line1_2(pnex, pney, pswx, pswy);
    FLineObj line1_3(pswx, pswy, pnwx, pnwy);
    
    // clockwise is better
    t2x[0] = pswx;
    t2y[0] = pswy;
    t2x[1] = pnex;
    t2y[1] = pney;
    t2x[2] = psex;
    t2x[2] = psey;
    FFillPolygonObj tri2(t2x, t2y, 3);
    FLineObj line2_1(pswx, pswy, pnex, pney);
    FLineObj line2_2(pnex, pney, psex, psey);
    FLineObj line2_3(psex, psey, pswx, pswy);

    double x1, y1, z1, x2, y2, z2, x3, y3, z3;
    double vert1[3], vert2[3], vert3[3], normal[3], probept[3];
    if (tri2.Contains(ppt) || line2_1.Contains(ppt) ||
    	line2_2.Contains(ppt) || line2_3.Contains(ppt)) {
      //x1 = swx;
      //y1 = swy;
      projected_position(row, col, x1, y1);
      z1 = value(row, col);
      //x2 = sex;
      //y2 = sey;
      projected_position(row, col+1, x2, y2);
      z2 = value(row, col+1);
      //x3 = nex;
      //y3 = ney;
      projected_position(row+1, col+1, x3, y3);
      z3 = value(row+1, col+1);

    }
    else /*if (tri1.Contains(ppt) || line1_1.Contains(ppt) ||
	     line1_2.Contains(ppt) || line1_3.Contains(ppt))*/ {
      //x1 = swx;
      //y1 = swy;
      projected_position(row, col, x1, y1);
      z1 = value(row, col);
      //x2 = nex;
      //y2 = ney;
      z2 = value(row+1, col+1);
      projected_position(row+1, col+1, x2, y2);
      //x3 = nwx;
      //y3 = nwy;
      projected_position(row+1, col, x3, y3);
      z3 = value(row+1, col);

    }
    vert1[0] = x1;
    vert1[1] = y1;
    vert1[2] = z1;
    vert2[0] = x2;
    vert2[1] = y2;
    vert2[2] = z2;
    vert3[0] = x3;
    vert3[1] = y3;
    vert3[2] = z3;
    probept[0] = px;
    probept[1] = py;
    TriangleComputeNormal(vert1, vert2, vert3, normal);
    v = PlaneEvaluateZ(normal, vert1, probept);

    delete [] t1x;
    delete [] t1y;
    delete [] t2x;
    delete [] t2y;
  }
  return 1;
}

int DEM::tri3d_by_utmpos(double ux, double uy,
			 double& x1, double& y1, double& z1,
			 double& x2, double& y2, double& z2,
			 double& x3, double& y3, double& z3
			 )
{
  if (_proj) {
    UV data;
    data.u = ux;
    data.v = uy;
    data = pj_inv(data, _proj);
    return tri3d_by_geopos(data.u * RAD_TO_DEG, data.v * RAD_TO_DEG,
			   x1, y1, z1, x2, y2, z2, x3, y3, z3);
  }
  else
    return 0;
}

int DEM::tri3d_by_geopos(double x, double y,
			 double& x1, double& y1, double& z1,
			 double& x2, double& y2, double& z2,
			 double& x3, double& y3, double& z3
			 )
{
  // first do a containment test in geo coords
  FPointObj pt(x, y);
  if (grunits == arcseconds) {
    FBoxObj box(_swe, _swn, _nee, _nen);
    if (!box.Contains(pt))
      return 0;
  }
  if (grunits == arcseconds) {
    int coli = 0;
    int rowi = 0;
    double dx, dy;
    for (coli = 0; coli < cols(); coli++) {
      position(0, coli, dx, dy);
      if (dx >= x)
	break;
    }
    for (rowi = 0; rowi < rows(); rowi++) {
      position(rowi, 0, dx, dy);
      if (dy >= y)
	break;
    }
    int col = coli - 1;
    int row = rowi - 1;

    double swx, swy, nwx, nwy, nex, ney, sex, sey;
    double pswx, pswy, pnwx, pnwy, pnex, pney, psex, psey;
    position(row, col, swx, swy);
    position(row+1, col, nwx, nwy);
    position(row+1, col+1, nex, ney);
    position(row, col+1, sex, sey);
    projected_position(row, col, pswx, pswy);
    projected_position(row+1, col, pnwx, pnwy);
    projected_position(row+1, col+1, pnex, pney);
    projected_position(row, col+1, psex, psey);
    UV data;
    double px, py;
    data.u = x * DEG_TO_RAD;
    data.v = y * DEG_TO_RAD;
    data = pj_fwd(data, _proj);
    px = data.u;
    py = data.v;
    FPointObj ppt(px, py);

    float* t1x = new float[3];
    float* t1y = new float[3];
    float* t2x = new float[3];
    float* t2y = new float[3];
    t1x[0] = pnwx;
    t1y[0] = pnwy;
    t1x[1] = pnex;
    t1y[1] = pney;
    t1x[2] = pswx;
    t1x[2] = pswy;
    FFillPolygonObj tri1(t1x, t1y, 3);
    FLineObj line1_1(pnwx, pnwy, pnex, pney);
    FLineObj line1_2(pnex, pney, pswx, pswy);
    FLineObj line1_3(pswx, pswy, pnwx, pnwy);
    
    // clockwise is better
    t2x[0] = pswx;
    t2y[0] = pswy;
    t2x[1] = pnex;
    t2y[1] = pney;
    t2x[2] = psex;
    t2x[2] = psey;
    FFillPolygonObj tri2(t2x, t2y, 3);
    FLineObj line2_1(pswx, pswy, pnex, pney);
    FLineObj line2_2(pnex, pney, psex, psey);
    FLineObj line2_3(psex, psey, pswx, pswy);

    if (tri2.Contains(ppt) || line2_1.Contains(ppt) ||
    	line2_2.Contains(ppt) || line2_3.Contains(ppt)) {
      //x1 = swx;
      //y1 = swy;
      projected_position(row, col, x1, y1);
      z1 = value(row, col);
      //x2 = sex;
      //y2 = sey;
      projected_position(row, col+1, x2, y2);
      z2 = value(row, col+1);
      //x3 = nex;
      //y3 = ney;
      projected_position(row+1, col+1, x3, y3);
      z3 = value(row+1, col+1);

    delete [] t1x;
    delete [] t1y;
    delete [] t2x;
    delete [] t2y;
      return 1;
    }
    else /*if (tri1.Contains(ppt) || line1_1.Contains(ppt) ||
	     line1_2.Contains(ppt) || line1_3.Contains(ppt))*/ {
      //x1 = swx;
      //y1 = swy;
      projected_position(row, col, x1, y1);
      z1 = value(row, col);
      //x2 = nex;
      //y2 = ney;
      z2 = value(row+1, col+1);
      projected_position(row+1, col+1, x2, y2);
      //x3 = nwx;
      //y3 = nwy;
      projected_position(row+1, col, x3, y3);
      z3 = value(row+1, col);

    delete [] t1x;
    delete [] t1y;
    delete [] t2x;
    delete [] t2y;
      return 1;
    }
    //else {
    //printf("oops\n");
    //return 0;
    //}
    delete [] t1x;
    delete [] t1y;
    delete [] t2x;
    delete [] t2y;
    return 0;
  }
  return 0;
}

int DEM::trimesh_by_geopos(double x0, double y0, double x1, double y1,
			   int& nt, double*& tx, double*& ty, double*& tz) {
  // first do a containment test in geo coords
  FBoxObj qbox(x0, y0, x1, y1);
  if (grunits == arcseconds) {
    FBoxObj box(_swe, _swn, _nee, _nen);
    if (!box.Intersects(qbox))
      return 0;
  }
  if (grunits == arcseconds) {
    int coli = 0;
    int rowi = 0;
    double dx, dy;
    for (coli = 0; coli < cols(); coli++) {
      position(0, coli, dx, dy);
      if (dx >= x0)
	break;
    }
    for (rowi = 0; rowi < rows(); rowi++) {
      position(rowi, 0, dx, dy);
      if (dy >= y0)
	break;
    }
    int col0 = coli - 1;
    int row0 = rowi - 1;
  
    for (coli = 0; coli < cols(); coli++) {
      position(0, coli, dx, dy);
      if (dx >= x1)
	break;
    }
    for (rowi = 0; rowi < rows(); rowi++) {
      position(rowi, 0, dx, dy);
      if (dy >= y1)
	break;
    }
    int col1 = coli;
    int row1 = rowi;
  
    if (row0 == row1 && col0 == col1)
      nt = 2 * 3;
    else if (row0 == row1)
      nt = (col1 - col0) * 2 * 3;
    else if (col0 == col1)
      nt = (row1 - row0) * 2 * 3;
    else
      nt = (((col1 - col0) * (row1 - row0)) * 2) * 3;
    tx = new double[nt];
    ty = new double[nt];
    tz = new double[nt];

    int ti = 0;
    for (coli = col0; coli < col1; coli++) {
      for (rowi = row0; rowi < row1; rowi++) {

	projected_position(rowi, coli, tx[ti], ty[ti]);
	tz[ti++] = value(rowi, coli);
	projected_position(rowi, coli+1, tx[ti], ty[ti]);
	tz[ti++] = value(rowi, coli+1);
	projected_position(rowi+1, coli+1, tx[ti], ty[ti]);
	tz[ti++] = value(rowi+1, coli+1);

	projected_position(rowi, coli, tx[ti], ty[ti]);
	tz[ti++] = value(rowi, coli);
	projected_position(rowi+1, coli+1, tx[ti], ty[ti]);
	tz[ti++] = value(rowi+1, coli+1);
	projected_position(rowi+1, coli, tx[ti], ty[ti]);
	tz[ti++] = value(rowi+1, coli);
      }
    }
    return 1;
  }
  return 0;
}

int DEM::trimesh_by_utmpos(double x0, double y0, double x1, double y1,
			   int& nt, double*& tx, double*& ty, double*& tz) {
  if (_proj) {
    double px0, py0, px1, py1;
    UV data;
    data.u = x0;
    data.v = y0;
    data = pj_inv(data, _proj);
    px0 = data.u * RAD_TO_DEG;
    py0 = data.v * RAD_TO_DEG;
    data.u = x1;
    data.v = y1;
    data = pj_inv(data, _proj);
    px1 = data.u * RAD_TO_DEG;
    py1 = data.v * RAD_TO_DEG;
    return trimesh_by_geopos(px0, py0, px1, py1, nt, tx, ty, tz);
  }
  else
    return 0;
}

void DEM::position(int row, int col, double& x, double& y) {
    int ro, co;
    if (row < 0)
	ro = 0;
    else if (row >= r)
	ro = r-1;
    else
	ro = row;
    if (col < 0)
	co = 0;
    else if (col >= c)
	co = c-1;
    else
	co = col;
    if (grunits == arcseconds) {
    x = _swe + (co * xres)/3600.0;
    y = _swn + (ro * yres)/3600.0;
    }
    else if (grunits == meters) {
    x = _swe + (co * xres);
    y = _swn + (ro * yres);
    }
}

void DEM::projected_position(int row, int col, double& x, double& y) {
    UV data;
    position(row, col, data.u, data.v);
    data.u *= DEG_TO_RAD;
    data.v *= DEG_TO_RAD;
    if (_proj) {
	data = pj_fwd(data, _proj);
	x = data.u;
	y = data.v;
    }
}

void DEM::projection(PJ* p) {
    _proj = p;
}

void DEM::xyzdump(FILE* f) {
    double x, y;
    for (int i = 0; i < cols(); i++) {
	for (int j = 0; j < rows(); j++) {
	    projected_position(j, i, x, y);
	    fprintf(f, "%f %f %d\n", x, y, value(j, i)); 
	}
    }
}

int DEM::geowithin(double x, double y) {
  //if (horizontal_units() == arcseconds) {
  if ((x >= _swe || x >= _nwe) && (x <= _nee || x <= _see) &&
      (y >= _swn || y >= _sen) && (y <= _nen || y <= _nwn))
    return 1;
  else
    return 0;
#if 0
  }
  else if (horizontal_units() == meters) {
  if ((x >= _swe-xres || x >= _nwe-xres) && (x <= _nee+xres || x <= _see+xres) &&
      (y >= _swn-yres || y >= _sen-yres) && (y <= _nen+yres || y <= _nwn+yres))
    return 1;
  else
    return 0;
  }
#endif
}
