/************************************************************************
 *   MODULE  : LOD_DTED.C
 *
 *   PURPOSE : contains routines to load and free the dted for a map
 *
 *   AUTHOR  : Ron Hawkins, SAIC (original)
 *             Kevin Russo, SAIC (1, 2, 4)
 *             Melissa Mathews, SAIC (3)
 *
 *   Change     Date         Reason
 *   ------     ----          ------
 *   Original   25 Sep 92    n/a
 *   1          20 Oct 92    overall program update
 *   2          01 Apr 92    update to XVT 3.1 and MUSE Release 3.0
 *   3          17 Jun 93    converted xvt 2.0 calls to xvt 3.1
 *   4          25 Jun 93    added routines to bypass those which don't
 *                           work in the new MUSE libraries
 *
 ************************************************************************/

/****************
 * header files *
 ****************/

#include "los.h"

/*********************************
 * file scope function prototype *
 *********************************/

#ifdef PROTO
#if XVT_OS == XVT_OS_WIN
int dted_construct (HFILE fhandle, RASTER *raster);
#else
int dted_construct (FILE *file, RASTER *raster);
#endif
#else
int dted_construct ();
#endif

/****************
 * declarations *
 ****************/

#define SZ_RASTER      1024

#if 0
/************************************************************************
 *   FUNCTION     : load_dted
 *
 *   PURPOSE      : calls the necessary functions to read the dted into
 *                  memory.
 *
 *   INPUTS       : none
 *
 *   RETURN VALUE : none
 *
 *   CALLED FROM  : do_control in module LOSW502.C
 *
 *   CALLS        : check_dted_exists in module LOD_DTED.C
 *                  dted_construct in module LOD_DTED.C
 *                  BoxWithin in module LOSMISC.C
 *                  available_structure {MUSE LIB}
 *                  status_update in module LOSMISC.C
 *                  dted_destruct in module LOD_DTED.C
 *
 ************************************************************************/

#ifdef PROTO
ERRSTATUS load_dted (void)
#else
ERRSTATUS load_dted ()
#endif
{
   AVAILABLE        *av=NULL;
   MAP_DOC          *mapdoc;
   BASEMAP          *basemap;
   BOOLEAN          no_dted_regions;
   EXTENT           box1, box2;

   int              status;

#if XVT_OS == XVT_OS_WIN
   HFILE            fhandle;
#else
   FILE             *fptr;
#endif

   extern RASTER    elev_raster;
   extern MAP_TYPE  mapinfo;

   /* get a pointer to the MUSE structure */
   available_structure(&av);
   if (av->windows->current_map == (WINDOW)NULL)
      xvt_dm_post_fatal_exit("No map is open");

   /* get the mapdoc and basemap pointers */
   mapdoc = (MAP_DOC *)xvt_vobj_get_data(av->windows->current_map);
   if (mapdoc == (MAP_DOC *)NULL)
      xvt_dm_post_fatal_exit("Unable to access the basemap map document");

   basemap = (BASEMAP *)mapdoc->products->basemap;
   if (basemap == (BASEMAP *)NULL)
      xvt_dm_post_fatal_exit("Unable to access the basemap data");

   /* check for dted file existance */
   if (check_dted_file_exists() == FALSE)
   {
      status_update ("\0");
      /*      xvt_dm_post_error("Unable to access the DTED file");*/
      return STAT_FILE_NOT_FOUND;
   }

   /* change to the dted file directory and load dted */
   xvt_fsys_set_dir(&mapinfo.dted_file.dir);

#if XVT_OS == XVT_OS_WIN
   fhandle = _lopen (mapinfo.dted_file.name, READ | OF_SHARE_EXCLUSIVE);
   if (fhandle == (HFILE)HFILE_ERROR)
   {
      status_update ("\0");
      /*      xvt_dm_post_error("Unable to open the DTED file");*/
      return STAT_FILE_NOT_FOUND;
   }
#else
   fptr = fopen(mapinfo.dted_file.name, "rb");
   if (fptr == (FILE *)NULL)
   {
      status_update ("\0");
      /*      xvt_dm_post_error("Unable to open the DTED file");*/
      return STAT_FILE_NOT_FOUND;
   }
#endif

   /* set the status message before loading */
   status_update ("Loading DTED.....");

#if XVT_OS == XVT_OS_WIN
   status = dted_construct (fhandle, &elev_raster);
   if (status != (int)0)
   {
      _lclose (fhandle);
      status_update ("\0");
      dted_destruct(&elev_raster);
      return status;
   }

   _lclose (fhandle);
#else
   status = dted_construct (fptr, &elev_raster);
   if (status != (int)0)
   {
      fclose (fptr);
      status_update ("\0");
      dted_destruct(&elev_raster);
      return status;
   }

   fclose (fptr);
#endif

   /* check DTED extents */
   if (!BoxWithin (elev_raster.data_hdr.extent, basemap->area->data_hdr.extent))
   {
      status_update ("\0");
#if 0
      xvt_dm_post_error("DTED lies outside map geographic boundaries");
      xvt_app_process_pending_events();
#endif
      dted_destruct (&elev_raster);
      mapinfo.dted_file.dir = basemap->file_spec.dir;
      strcpy(mapinfo.dted_file.name, basemap->file_spec.name);
      strcpy(mapinfo.dted_file.type, FT_RAS_DTED);
      mapinfo.dted_file.name[strlen(mapinfo.dted_file.name)-4] = '\0';
      strcat(mapinfo.dted_file.name, ".dte");
      return STAT_ABORT;
   }

   /* check for no DTED regions on map */
   box1 = elev_raster.data_hdr.extent;
   box2 = basemap->area->data_hdr.extent;

   no_dted_regions = FALSE;

   if ((box1.left > box2.left) || (box1.right < box2.right))
      no_dted_regions = TRUE;

   if ((box1.bottom > box2.bottom) || (box1.top < box2.top))
      no_dted_regions = TRUE;

   if (no_dted_regions)
   {
      status_update ("\0");
#if 0
      xvt_dm_post_note("Some basemap regions do not contain DTED");
      xvt_app_process_pending_events();
#endif
   }

   /* reset the message status */
   status_update ("\0");

   return STAT_SUCCESS;
}

/************************************************************************
 *   FUNCTION     : check_dted_file_exists
 *
 *   PURPOSE      : checks if the mapinfo dted file name structure
 *                  contains a valid dted file name by trying to access
 *                  the dted file without actually opening it.
 *
 *   INPUTS       : void
 *
 *   RETURN VALUE : BOOLEAN - FALSE on error, TRUE otherwise
 *
 *   CALLED FROM  : load_dted in module LOD_DTED.C
 *
 *   CALLS        : check_file_name in module LOSMISC.C
 *
 ************************************************************************/

#ifdef PROTO
BOOLEAN check_dted_file_exists(void)
#else
BOOLEAN check_dted_file_exists()
#endif
{
   FILE_SPEC        temp_file_spec;

   extern MAP_TYPE  mapinfo;

   xvt_fsys_set_dir(&mapinfo.dted_file.dir);

   if (access(mapinfo.dted_file.name, 0) != 0)
   {
      temp_file_spec.dir = mapinfo.dted_file.dir;
      strcpy(temp_file_spec.name, mapinfo.dted_file.name);
#if XVT_OS == XVT_OS_MAC
      strcpy(temp_file_spec.type, "\0");
#else
      strcpy(temp_file_spec.type, FT_RAS_DTED);
#endif
      switch (xvt_dm_post_file_open(&temp_file_spec, "Please locate DTED file..."))
      {
         case FL_BAD:
         case FL_CANCEL:
            return FALSE;
         case FL_OK:
            if (check_file_name (temp_file_spec.name, "DTE") == FALSE)
               return FALSE;
            mapinfo.dted_file = temp_file_spec;
            break;
      }
   }

   return TRUE;
}
#endif

/************************************************************************
 *   FUNCTION     : dted_destruct
 *
 *   PURPOSE      : free the dted data
 *
 *   INPUTS       : raster - pointer to the elev_raster structure
 *
 *   RETURN VALUE : none
 *
 *   CALLED FROM  : load_dted in module LOD_DTED.C
 *                  do_file_menu in module LOSMENU1.C
 *
 *   CALLS        : none
 *
 ************************************************************************/

#ifdef PROTO
void dted_destruct(RASTER *raster)
#else
void dted_destruct(raster)
RASTER *raster;
#endif
{
   BOOLEAN status;

   extern RASTER elev_raster;

   if (raster->data_matrix.ghandle != (GHANDLE)NULL)
   {
      status = xvt_gmem_free (raster->data_matrix.ghandle);
      if (status == FALSE)
	/*         xvt_dm_post_error("Unable to release dted data memory")*/;

   }

   raster->data_matrix.ghandle = (GHANDLE)NULL;
}

/************************************************************************
 *   FUNCTION     : dted_construct
 *
 *   PURPOSE      : load the dted into memory
 *
 *   INPUTS       : file   - file pointer (handle under WINDOWS)
 *                  raster - pointer to a raster structure
 *
 *   RETURN VALUE : 0 on success, -1 otherwise
 *
 *   CALLED FROM  : multiple places in the LOS code
 *
 *   CALLS        : decode_raster_hdr {MUSE LIB},
 *                  big_endian_machine in mocule SWAP.C
 ************************************************************************/

#ifdef PROTO
#if XVT_OS == XVT_OS_WIN
int dted_construct(HFILE fhandle, RASTER *raster)
#else
int dted_construct(FILE *file, RASTER *raster)
#endif
#else
#if XVT_OS == XVT_OS_WIN
int dted_construct(fhandle, raster)
HFILE fhandle;
RASTER *raster;
#else
int dted_construct(file, raster)
FILE *file;
RASTER *raster;
#endif
#endif
{
   size_t items_read;

   char pHMEM data;

   short pHMEM dted;

   long loop, width, height, data_read;

   unsigned long length_in_bytes;
   unsigned char hdr[SZ_RASTER];

   /* read the DTED raster file header */
#if XVT_OS == XVT_OS_WIN
   items_read = (size_t)_lread(fhandle, (void pHMEM)hdr, (long)SZ_RASTER);

   /* check the header for errors */
   if (items_read != SZ_RASTER)
   {
      xvt_dm_post_error("Unable to read in the raster DTED");
      return (int)-1;
   }
#else
   items_read = fread((void *)hdr, sizeof(hdr), (size_t)1, file);

   /* check the header for errors */
   if (items_read != 1)
   {
     /*      xvt_dm_post_error("Unable to read in the raster DTED");*/
      return (int)-1;
   }
#endif

   decode_raster_hdr(hdr, raster);

   if (raster->magic != MAGIC_RASTER)
   {
     /*      xvt_dm_post_error("The raster data read in is incorrect.");*/
      return (int)-1;
   }

   /* set the palette and LUT pointers */
   raster->palette = (PALETTE *)NULL;
   raster->lut     = (LUT *)NULL;

   /* calculate the data matrix size */
   length_in_bytes = (raster->data_matrix.width * raster->data_matrix.height
                      * raster->data_matrix.bits_per_pixel) / BITS_PER_BYTE;

   /* allocate the raster memory */
   raster->data_matrix.ghandle = xvt_gmem_alloc(length_in_bytes);
   if (raster->data_matrix.ghandle == (GHANDLE)NULL)
      return STAT_MEM_ALLOC_ERR;

   /* read the data */
   data = (char pHMEM)xvt_gmem_lock(raster->data_matrix.ghandle);
   if (data == (char pHMEM)NULL)
      xvt_dm_post_fatal_exit("Unable to lock down DTED matrix");

#if XVT_OS == XVT_OS_WIN
   data_read = _hread (fhandle, (void pHMEM)data, (long)length_in_bytes);

   /* check the header for errors */
   if (data_read != length_in_bytes)
   {
     /*      xvt_dm_post_error("Unable to read in the raster DTED");*/
      return (int)-1;
   }
#else
   items_read = fread (data, length_in_bytes * sizeof(char), 1, file);

   /* check the header for errors */
   if (items_read != 1)
   {
     /*      xvt_dm_post_error("Unable to read in the raster DTED");*/
      return (int)-1;
   }
#endif

   xvt_gmem_unlock (raster->data_matrix.ghandle);

   if (big_endian_machine())
   {
      dted = (short pHMEM)xvt_gmem_lock(raster->data_matrix.ghandle);
      if (dted == (short pHMEM)NULL)
         xvt_dm_post_fatal_exit("Unable to lock down DTED matrix");

      width = raster->data_matrix.width;

      height = raster->data_matrix.height;

      for (loop = 0; loop < (width * height); loop++)
          dted[loop] = swap_short(dted[loop]);

      xvt_gmem_unlock (raster->data_matrix.ghandle);
   }

   return (int)0;
}
